use super::position::Span;



/// The different types of tokens that the `Lexer` will break a
/// file down into. These are all base token types. It is left as
/// basic as it is so that they can be combined and altered for
/// each domains specific needs.
///
/// Base tokens are the most basic token type.
/// Complex tokens are a token type that uses multiple base tokens.
/// Swap tokens are when you swap one token type into another.
#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
pub enum TokenType
{
   /// This represents a single newline character.
   ///
   /// This is a base token type.
   Newline,

   /// A numeric character followed by zero or more numeric characters.
   ///
   /// This is a base token type.
   Numeric,

   /// A non alphanumeric or whitespace character.
   ///
   /// This is a base token type.
   Symbol,

   /// An alpha character followed by zero or more alpha characters.
   ///
   /// This is a base token type.
   Text,

   /// One or more whitespace characters.
   ///
   /// This is a base token type.
   Whitespace,

   /// This token has something, but we don't know what.
   /// This should not occur.
   Unknown
}



/// Data oriented design of the information desired while
/// tokenizing a file.
///
/// Everything is in flat arrays for fast access
/// and minimal cache misses.
#[derive(Debug, Clone, Default)]
pub struct TokenStream<T>
{
   /// The text of the `Token`.
   pub lexemes: Vec<String>,

   /// The type of `Token`.
   pub variants: Vec<T>,

   /// The location of the `Token` in the file.
   pub locations: Vec<Span>
}


/// A `Token` found during the lexigraphical scan.
#[derive(Debug)]
pub struct Token<'a, T>
{
   /// The characters of the `Token`.
   pub lexeme: &'a str,

   /// The `Token`'s type.
   pub variant: &'a T,

   /// The location in the file of this `Token`.
   pub span: &'a Span
}

/// A `Token` found during the lexigraphical scan.
///
/// This is the mutable reference.
#[derive(Debug)]
pub struct TokenMut<'a, T>
{
   /// The characters of the `Token`.
   pub lexeme: &'a mut str,
   /// The `Token`'s type.
   pub variant: &'a mut T,
   /// The location for this `Token` in the file.
   pub span: &'a mut Span
}


/// An iterator over the Tokens in a `TokenStream`.
pub struct TokenStreamIter<'a, T>
{
   /// The stream to iterate over.
   stream: &'a TokenStream<T>,

   /// The position in the stream.
   index: usize
}

/// A mutable iterator over the Tokens in a `TokenStream`.
pub struct TokenStreamIterMut<'a, T>
{
   /// The characters of the `Token`.
   lexemes: std::slice::IterMut<'a, String>,

   /// The `Token`'s type.
   variants: std::slice::IterMut<'a, T>,

   /// The location for this `Token` in the file.
   locations: std::slice::IterMut<'a, Span>
}



impl<T> TokenStream<T>
{
   pub fn new() -> Self
   {
      TokenStream { lexemes: Vec::new(),
                    variants: Vec::new(),
                    locations: Vec::new() }
   }

   pub fn len(&self) -> usize
   {
      self.lexemes.len()
   }

   pub fn is_empty(&self) -> bool
   {
      self.lexemes.is_empty()
   }

   pub fn get(&self, index: usize) -> Option<Token<'_, T>>
   {
      if index < self.lexemes.len()
      {
         Some(Token { lexeme: &self.lexemes[index],
                      variant: &self.variants[index],
                      span: &self.locations[index] })
      }
      else
      {
         None
      }
   }

   pub fn iter(&self) -> TokenStreamIter<'_, T>
   {
      TokenStreamIter { stream: self,
                        index: 0 }
   }

   pub fn get_mut(&mut self, index: usize) -> Option<TokenMut<'_, T>>
   {
      if index < self.lexemes.len()
      {
         // SAFETY: We manually split the borrows to avoid
         // double mutable borrow.
         let lexeme = &mut self.lexemes[index] as *mut String;
         let variant = &mut self.variants[index] as *mut T;
         let span = &mut self.locations[index] as *mut Span;

         // Convert &mut String to &mut str safely.
         unsafe {
            Some(TokenMut { lexeme: &mut *lexeme.as_mut()
                                                .unwrap()
                                                .as_mut_str(),
                            variant: &mut *variant,
                            span: &mut *span })
         }
      }
      else
      {
         None
      }
   }

   pub fn clear(&mut self)
   {
      self.lexemes.clear();
      self.variants.clear();
      self.locations.clear();
   }

   pub fn push(&mut self, lexeme: String, variant: T, span: Span)
   {
      self.lexemes.push(lexeme);
      self.variants.push(variant);
      self.locations.push(span);
   }

   pub fn iter_mut(&mut self) -> TokenStreamIterMut<'_, T>
   {
      TokenStreamIterMut { lexemes: self.lexemes.iter_mut(),
                           variants: self.variants.iter_mut(),
                           locations: self.locations.iter_mut() }
   }
}


impl<'a, T> IntoIterator for &'a TokenStream<T>
{
   type IntoIter = TokenStreamIter<'a, T>;
   type Item = Token<'a, T>;

   fn into_iter(self) -> Self::IntoIter
   {
      TokenStreamIter { stream: self,
                        index: 0 }
   }
}

impl<'a, T> Iterator for TokenStreamIter<'a, T>
{
   type Item = Token<'a, T>;

   fn next(&mut self) -> Option<Self::Item>
   {
      if self.index < self.stream.lexemes.len()
      {
         let i = self.index;
         self.index += 1;

         Some(Token { lexeme: &self.stream.lexemes[i],
                      variant: &self.stream.variants[i],
                      span: &self.stream.locations[i] })
      }
      else
      {
         None
      }
   }
}


impl<'a, T> Iterator for TokenStreamIterMut<'a, T>
{
   type Item = TokenMut<'a, T>;

   fn next(&mut self) -> Option<Self::Item>
   {
      let lexeme = self.lexemes.next()?;
      let variant = self.variants.next()?;
      let span = self.locations.next()?;

      Some(TokenMut { lexeme: &mut *lexeme,
                      variant,
                      span })
   }
}


impl<'a, T: std::fmt::Display> ::std::fmt::Display for Token<'a, T>
{
   fn fmt(&self, f: &mut ::std::fmt::Formatter) -> ::std::fmt::Result
   {
      match self.lexeme
      {
         "\n" => write!(f, "[{}, {}]", self.variant, "\\n"),

         _ => write!(f, "[{}: {}]", self.variant, self.lexeme)
      }
   }
}

impl ::std::fmt::Display for TokenType
{
   fn fmt(&self, f: &mut ::std::fmt::Formatter) -> ::std::fmt::Result
   {
      let name = match *self
      {
         TokenType::Newline => "Newline",

         TokenType::Numeric => "Numeric",

         TokenType::Symbol => "Symbol",

         TokenType::Text => "Text",

         TokenType::Whitespace => "Whitespace",

         TokenType::Unknown => "Unknown"
      };

      write!(f, "{}", name)
   }
}
